/* ReVivAl v0.16 */
/* by Cedric Piette */

/* based on */
/**************************************************************************************************
MiniSat 
http://www.cs.chalmers.se/Cs/Research/FormalMethods/MiniSat/
**************************************************************************************************/


/* *************************** */
/*    OPTIONS                  */
/* (un)comment to (un)activate */
/* *************************** */

//#define RANDOMIZED_SEED
//#define PRINT_DETAILS




#include "solver.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <zlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>

#define MAX_NUMBER_CLAUSE 10000000

int i,globalModif,lastModif;


//=================================================================================================
// Helpers:


// Reads an input stream to end-of-file and returns the result as a 'char*' terminated by '\0'
// (dynamic allocation in case 'in' is standard input).
//
char* readFileMin(FILE *  in)
{
    char*   data = malloc(65536);
    int     cap  = 65536;
    int     size = 0;

    while (!feof(in)){
        if (size == cap){
            cap *= 2;
            data = realloc(data, cap); }
        size += fread(&data[size], 1, 65536, in);
    }
    data = realloc(data, size+1);
    data[size] = '\0';

    return data;
}

//static inline double cpuTime(void) {
//    struct rusage ru;
//    getrusage(RUSAGE_SELF, &ru);
//    return (double)ru.ru_utime.tv_sec + (double)ru.ru_utime.tv_usec / 1000000; }


//=================================================================================================
// DIMACS Parser:


static inline void skipWhitespaceMin(char** in) {
    while ((**in >= 9 && **in <= 13) || **in == 32)
        (*in)++; }

static inline void skipLineMin(char** in) {
    for (;;){
        if (**in == 0) return;
        if (**in == '\n') { (*in)++; return; }
        (*in)++; } }

static inline int parseIntMin(char** in) {
    int     val = 0;
    int    _neg = 0;
    skipWhitespaceMin(in);
    if      (**in == '-') _neg = 1, (*in)++;
    else if (**in == '+') (*in)++;
    if (**in < '0' || **in > '9') fprintf(stderr, "PARSE ERROR! Unexpected char: %c\n", **in), exit(1);
    while (**in >= '0' && **in <= '9')
        val = val*10 + (**in - '0'),
        (*in)++;
    return _neg ? -val : val; }

static void readClauseMin(char** in, solver* s, veci* lits) {
    int parsed_lit, var;
    veci_resize(lits,0);
    for (;;){
        parsed_lit = parseIntMin(in);
        if (parsed_lit == 0) break;
        var = abs(parsed_lit)-1;
        veci_push(lits, (parsed_lit > 0 ? toLit(var) : lit_neg(toLit(var))));
		//printf("Litteral : %d -- TRANSFO minisat : %d\n",parsed_lit,(parsed_lit > 0 ? toLit(var) : lit_neg(toLit(var))));
    }
}

static lbool parse_DIMACS_mainMin(char* in, solver* s) {
    veci lits;
    veci_new(&lits);

    for (;;){
        skipWhitespaceMin(&in);
        if (*in == 0)
            break;
        else if (*in == 'c' || *in == 'p')
            skipLineMin(&in);
        else{
            lit* begin;
            readClauseMin(&in, s, &lits);
            begin = veci_begin(&lits);
            if (!solver_addclause(s, begin, begin+veci_size(&lits))){
                veci_delete(&lits);
                return l_False;
            }
        }
    }
    veci_delete(&lits);
    return solver_simplify(s);
}


// Inserts problem into solver. Returns FALSE upon immediate conflict.
//
static lbool parse_DIMACSMin(FILE * in, solver* s) {
    char* text = readFileMin(in);
    lbool ret  = parse_DIMACS_mainMin(text, s);
    free(text);
    return ret; }


//=================================================================================================

void printStats(stats* stats, int cpu_time){
  return ;
}

/* FOR PRINTING THE CLAUSES.
	for(i=0;i<s->stats.clauses;i++){
		currentClause=(vecp_begin(&(s->clauses)))[i];
		size=clause_size(currentClause);
		for(j=0;j<size;j++){
			printf("%d ",clause_giveLit(currentClause,j));
		}
		printf("\n");
	}*/

/* FOR PRINTING THE CURRENT ASSIGNMENT. 
		for(j=0;j<s->size;j++){
			if(!AmIUndef(j<<1))
				printf("%d=%c  ",j+1,s->assigns[j]==l_False?'F':'T');
		}
		printf("\n");*/


#define MAX_SIZE_LEARNT_CLAUSE 4

#define wellWritten(lit) (lit&1?(-(lit_var(lit)+1)):(lit_var(lit)+1))

static inline void wellWriteClause(clause *c){
  int i,size=clause_size(c);
  for(i=0;i<size;i++) printf("%d ",wellWritten(clause_giveLit(c,i)));
  printf("\n");
}

static inline void wellWriteClause_stderr(clause *c){
  int i,size=clause_size(c);
  for(i=0;i<size;i++) fprintf(stderr,"%d ",wellWritten(clause_giveLit(c,i)));
}

static inline void s_removeClause(solver *s, clause *c){
  lit* lits = clause_begin(c);
  vecp_remove(solver_read_wlist(s,lit_neg(lits[0])),(void*)(clause_size(c) > 2 ? c : clause_from_lit(lits[1])));
  vecp_remove(solver_read_wlist(s,lit_neg(lits[1])),(void*)(clause_size(c) > 2 ? c : clause_from_lit(lits[0])));
}

static inline void s_addItAgain(solver *s, clause *c){
  vecp_push(solver_read_wlist(s,lit_neg(c->lits[0])),(void*)(clause_size(c) > 2 ? c : clause_from_lit(c->lits[1])));
  vecp_push(solver_read_wlist(s,lit_neg(c->lits[1])),(void*)(clause_size(c) > 2 ? c : clause_from_lit(c->lits[0])));
}

#define simulate_removeClause(s,c)     s_removeClause(s,c),c=NULL
#define simulate_addItAgain(s,place,c) s_addItAgain(s,c),place=c
#define AmIUndef(lit) (s->assigns[lit_var(lit)]==l_Undef)
#define AmISat(lit) (((s->assigns[lit_var(lit)]==l_False) && (lit&1))  ||  ((s->assigns[lit_var(lit)]==l_True) && (!(lit&1))))
#define negation(lit) (lit^1)
#define free_clause(c) free(c),c=NULL
#define reverseLit(cl,num1,num2) useF1=clause_giveLit(currentClause,num1),clause_setLit(currentClause,num1,clause_giveLit(currentClause,num2)),clause_setLit(currentClause,num2,useF1);


#ifdef PRINT_DETAILS
#define opt_printf(arg) fprintf(stderr,arg);
#define opt_printf_num_current_clause() fprintf(stderr,"clause %d : ",i)
#define opt_printfclause(clause) wellWriteClause_stderr(clause)

#else
#define opt_printf(arg) 
#define opt_printf_num_current_clause()
#define opt_printfclause(clause) 

#endif

int timeOut=0;

void endVivification(){
  lastModif=globalModif=0;
  timeOut=1;
  return ;
}

int vivification(char* CNFfile, int mytime){
    solver* s = solver_new();
    lbool   st;
    FILE *  in;
    veci    learnt_clause;
    clause *currentClause,*learnt_cl=NULL,*c;  
    lit  *begin, *end;
    int j,k,l,size,lit,clause_sat,NotRedundant,change=0,FinalNumClauses,useF1,
      nbIteration=0,my_lit,num_assuming,num_unit=0,p1,p2;

#ifdef RANDOMIZED_ALGORITHM
    srand(time(NULL));
#else
    srand(1);
#endif

    if((in = fopen(CNFfile, "rb"))==NULL) fprintf(stderr, "ERROR! Could not open file: %s\n",CNFfile),exit(1);
    st = parse_DIMACSMin(in, s);
    fclose(in);
    if(MAX_NUMBER_CLAUSE<s->stats.clauses)
      fprintf(stderr,"ReVivAl: too many clauses ! (%d, MAX is %d)\n",(int) s->stats.clauses,MAX_NUMBER_CLAUSE),
      exit(-1);
    alarm(mytime);
    signal(SIGALRM,endVivification);
    
    FinalNumClauses=lastModif=s->stats.clauses;
    globalModif=1;
    while(globalModif){
      ++nbIteration;
      globalModif=0;
      for(i=0;i<(globalModif?s->stats.clauses:lastModif);i=((change)?(i):(i+1))){
	opt_printf_num_current_clause();
	if(!(currentClause=(vecp_begin(&(s->clauses)))[i])) { opt_printf("empty\n"); change=0; continue;}
	/* The considered clause is removed from the CNF */
	simulate_removeClause(s,(vecp_begin(&(s->clauses)))[i]);
	clause_sat=0,change=0,NotRedundant=1;
	if(solver_propagate(s)) {fprintf(stderr,"Proved UNSATISFIABLE by ReVivAl !!\n(polynomial instance)\n"); exit(0);}
	/* Clause analyzing. If it is already satisfied, then it is not considered. 
	   If it contains falsified literals, then they are removed. */
	for(j=0;j<clause_size(currentClause);j++){
	  if(!AmIUndef(clause_giveLit(currentClause,j))){
	    if(!AmISat(clause_giveLit(currentClause,j))){
	      for(k=(j+1);k<clause_size(currentClause);k++) 
		clause_begin(currentClause)[k-1]=clause_begin(currentClause)[k];
	      currentClause->size_learnt=((clause_size(currentClause))-1)<<1;
	      j-=1;
	    }
	    else{clause_sat=1; break;}
	  }
	}
	if(clause_sat) {
	  opt_printfclause(currentClause);
	  opt_printf("  Already SAT\n"); 
	  free_clause(currentClause); 
	  FinalNumClauses--; 
	  continue; 
	}
	if(clause_size(currentClause)==1) {
	  num_unit++;
	  enqueue(s,clause_giveLit(currentClause,0),(clause*)0); 
	  free_clause(currentClause); 
	  FinalNumClauses--; 
	  continue;
	}
	/* End of clause analyzing */

	/* Optional Printing of the original clause */
	opt_printfclause(currentClause);
	size=clause_size(currentClause);
	num_assuming=0;
	/* The MOMS-like sort of the clause has been removed */
	/* MOMS_sort(currentClause); */
	/* Two randomly selected literals are reversed */
	p1=random()%(clause_size(currentClause));
	p2=random()%(clause_size(currentClause));
	reverseLit(currentClause,p1,p2);
	/* Assignment of the opposite of each literal from the clause */
       	for(j=0;(j<size) && (NotRedundant) && (!change);j++){	  
	  if(clause_size(currentClause)==1) {break;}
	  lit=clause_giveLit(currentClause,j);
	  num_assuming++;
	  assume(s,negation(lit)); 
	  NotRedundant=((learnt_cl=solver_propagate(s))==NULL);	
	  for(k=j+1;k<size;k++){
	    my_lit=clause_giveLit(currentClause,k);
	    // The literal is assigned to the same value. the remaining part of the clause can be "omitted" 
	    if(s->assigns[lit_var(my_lit)]!=l_Undef){
	      if(AmISat(my_lit)){
		if(clause_size(currentClause)!=num_assuming+1){
		  change=1; 
		  clause_begin(currentClause)[num_assuming]=my_lit;
		  currentClause->size_learnt=(num_assuming+1)<<1;
		}else{
		  // In this case, the clause is redundant but no shorten clause can be deduced 
		  // (the last literal already assigned) 
		  j=size;
		  break;
		}
	      }
	      // The literal is assigned to the opposite value. it can be removed from the clause 
	      else{ 
		for(l=k+1;l<size;l++) 
		  clause_begin(currentClause)[l-1]=clause_giveLit(currentClause,l); 
		currentClause->size_learnt=(size-1)<<1; 
		change=1;
	      }
	      break;
	    }
	  }
	}
	if(!NotRedundant){
	  veci_new(&learnt_clause);
	  veci_resize(&learnt_clause,0);
	  solver_analyze(s, learnt_cl, &learnt_clause);
	  begin = veci_begin(&learnt_clause);
	  end   = begin + veci_size(&learnt_clause);
	  if(veci_size(&learnt_clause) > 1){
	    if(veci_size(&learnt_clause) < clause_size(currentClause)){
	      c     = clause_new(s,begin,end,1);
	      vecp_push(&s->learnts,c);
	      s->stats.learnts++;
	    }
	  }
	  else{
	    solver_canceluntil(s,s->root_level);
	    num_unit++;
	    enqueue(s,*begin,0);
	  }
	  if(clause_size(currentClause)!=num_assuming){
	    currentClause->size_learnt=(num_assuming)<<1,change=1;
	  }
	}
	solver_canceluntil(s,s->root_level);
	if(clause_size(currentClause)==1){ 
	  opt_printf(" --> "); opt_printfclause(currentClause); opt_printf("(unit)\n");
	  num_unit++;
	  enqueue(s,clause_giveLit(currentClause,0),(clause*)0); 
	  free_clause(currentClause); 
	  FinalNumClauses--; 
	  continue; 
	}
	simulate_addItAgain(s,(vecp_begin(&(s->clauses)))[i],currentClause);
	if(change){
	  if(!timeOut) {globalModif=1;  lastModif=i+1;} 
	  opt_printf(" --> "); opt_printfclause(currentClause);
	}
	opt_printf("\n");
      }
    }    
    if(solver_propagate(s)) {fprintf(stderr,"Proved UNSATISFIABLE by ReVivAl !!\n(polynomial instance)\n"); exit(0);}
    solver_canceluntil(s,s->root_level);
    
    fprintf(stderr,"#ITERATIONS %d\n",nbIteration);
    fprintf(stderr,"#LEARNT CLAUSES %d\n",(int) s->stats.learnts);
    
    /* Export produced CNF */
    /* Detected unit clauses first, then shortened clauses from the original CNF, then learnt clauses (3 loops) */
       
    printf("p cnf %d %d\n",s->size,(int) ((FinalNumClauses)+((int) s->stats.learnts)+num_unit));
    for(i=0,size=s->size;i<size;i++){
      if(s->assigns[i]!=l_Undef){
	printf("%d 0\n",s->assigns[i]==l_True?i+1:-(i+1));
      }      
    }
    
    for(i=0,size=s->stats.clauses;i<size;i++){
      currentClause=(vecp_begin(&(s->clauses)))[i];
      if(currentClause){
	k=clause_size(currentClause);
	for(j=0;j<k;j++) printf("%d ",wellWritten(clause_giveLit(currentClause,j)));
	printf("0\n");
      }
    }

    for(i=0,size=s->stats.learnts;i<size;i++){
      currentClause=(vecp_begin(&(s->learnts)))[i];
      if(currentClause){
	k=clause_size(currentClause);
	for(j=0;j<k;j++) printf("%d ",wellWritten(clause_giveLit(currentClause,j)));
	printf("0\n");
      }
    }
    return (int) st;
}


int main(int argc, char** argv){
  int mytime=0;
  switch(argc){
  case 3 : mytime=atoi(argv[2]);
  case 2 : break;
  default:  
    fprintf(stderr, "--- ReVivAl v0.16 ---\nusage: %s benchmark.cnf [time]\n",argv[0]),
    fprintf(stderr, "Vivification of the CNF formula 'benchmark.cnf' (DIMACS format)\n"),
    fprintf(stderr, "The resulting CNF is exported in stdout\n\nOption:\n"),
    fprintf(stderr, "time\t Stop the computation after 'time' seconds (no limit if not provided)\n"),
    exit(1);     
  }
  vivification(argv[1],mytime);
  exit(0);
}

